#include <param.h>
#include <systm.h>
#include <dir.h>
#include <signal.h>
#include <user.h>
#include <filsys.h>
#include <file.h>
#include <conf.h>
#include <inode.h>
#include <mount.h>
#include <reg.h>
#include <acct.h>
#include <errno.h>

extern struct file file[];
extern struct user u;
extern struct mount mount[];

/*
 * Convert a user supplied
 * file descriptor into a pointer
 * to a file structure.
 * Only task is to check range
 * of the descriptor.
 */
struct file *
getf(f)
register int f;
{
	register struct file *fp;

	if(0 <= f && f < NOFILE)
	{
		fp = u.u_ofile[f];
		if(fp != NULL)
			return((unsigned)fp & ~01);
	}
	u.u_error = EBADF;
	return(NULL);
}

/*
 * Internal form of close.
 * Decrement reference count on
 * file structure.
 * Also make sure the pipe protocol
 * does not constipate.
 *
 * Decrement reference count on the inode following
 * removal to the referencing file structure.
 * Call device handler on last close.
 */
closef(fp)
register struct file *fp;
{
	register struct inode *ip;
	register struct mount *mp;
	int flag, mode;
	dev_t dev;

	if(fp == NULL)
		return;
	ip = fp->f_inode;
	if (ip == u.u_locki)
		unlock();
	if (fp->f_count > 1)
	{
		fp->f_count--;
		return;
	}
	flag = fp->f_flag;
	dev = getdevno(ip);
	mode = ip->i_mode & IFMT;

	plock(ip);
	fp->f_count = 0;
	if(mode == IFIFO)
	{
		ip->i_mode &= ~(IREAD|IWRITE);
		wakeup((caddr_t)ip+1);
		wakeup((caddr_t)ip+2);
	}
	iput(ip);
	if (mode != IFCHR && mode != IFBLK)
		return;

	for(fp = file; fp < &file[NFILE]; fp++)
		if (fp->f_count && fp->f_inode == ip)
			return;
	if (mode == IFBLK)
	{
		for (mp = &mount[0]; mp < &mount[NMOUNT]; mp++)
			if (mp->m_bufp != NULL && mp->m_dev == dev)
				return;
		if (dev == swapdev)
			return;
		bflush(dev);
		(*bdevsw[major(dev)].d_close)(dev, flag);
		binval(dev);	/* devalidate associated blocks */
	}
	else
		(*cdevsw[major(dev)].d_close)(dev, flag);
}

/*
 * openi called to allow handler
 * of special files to initialize and
 * validate before actual IO.
 * flag is from fp->f_flag.
 */
openi(ip, flag)
register struct inode *ip;
{
	dev_t dev;
	register unsigned int maj;

	dev = getdevno(ip);
	maj = major(dev);
	switch(ip->i_mode & IFMT)
	{
	case IFCHR:
		if(maj >= nchrdev)
			goto bad;
		(*cdevsw[maj].d_open)(dev, flag);
		break;

	case IFBLK:
		if(maj >= nblkdev)
			goto bad;
		(*bdevsw[maj].d_open)(dev, flag);
	}
	return;

bad:
	u.u_error = ENXIO;
}

/*
 * Check mode permission on inode pointer.
 * Mode is READ, WRITE or EXEC.
 * In the case of WRITE, the
 * read-only status of the file
 * system is checked.
 * Also in WRITE, prototype text
 * segments cannot be written.
 * The mode is shifted to select
 * the owner/group/other fields.
 * The super user is granted all
 * permissions.
 */
access(ip, mode)
register struct inode *ip;
{
	register m;

	m = mode;
	if(m == IWRITE)
	{
		if(getfs(ip->i_dev)->s_ronly != 0)
		{
			u.u_error = EROFS;
			return(1);
		}
		if (ip->i_flag&ITEXT)		/* try to free text */
			xrele(ip);
		if(ip->i_flag & ITEXT)
		{
			u.u_error = ETXTBSY;
			return(1);
		}
	}
	if(u.u_uid == 0)
		return(0);
	if(u.u_uid != ip->i_uid)
	{
		m >>= 3;
		if(u.u_gid != ip->i_gid)
			m >>= 3;
	}
	if((ip->i_mode&m) != 0)
		return(0);

	u.u_error = EACCES;
	return(1);
}

/*
 * Look up a pathname and test if
 * the resultant inode is owned by the
 * current user.
 * If not, try for super-user.
 * If permission is granted,
 * return inode pointer.
 */
struct inode *
owner()
{
	register struct inode *ip;

	ip = namei(uchar, 0);
	if(ip == NULL)
		return(NULL);
	if(u.u_uid == ip->i_uid)
		return(ip);
	if(suser())
		return(ip);
	iput(ip);
	return(NULL);
}

/*
 * Test if the current user is the
 * super user.
 */
suser()
{

	if(u.u_uid == 0)
	{
		u.u_acflag |= ASU;
		return(1);
	}
	u.u_error = EPERM;
	return(0);
}

/*
 * Allocate a user file descriptor, starting from arg.
 */
ufalloc(i)
register i;
{
	while (i < NOFILE)
	{
		if(u.u_ofile[i] == NULL)
		{
			u.u_r.r_r0 = i;
			return(i);
		}
		i++;
	}
	u.u_error = EMFILE;
	return(-1);
}

/*
 * Allocate a user file descriptor
 * and a file structure.
 * Initialize the descriptor
 * to point at the file structure.
 *
 * no file -- if there are no available
 * 	file structures.
 */
struct file *
falloc()
{
	register struct file *fp;
	register i;

	i = ufalloc(0);
	if(i < 0)
		return(NULL);
	for(fp = &file[0]; fp < &file[NFILE]; fp++)
		if(fp->f_count == 0)
		{
			u.u_ofile[i] = fp;
			fp->f_count++;
			fp->f_offset = 0;
			return(fp);
		}
	printf("no file\n");
	uprints("\nOpen - File table full\n");
	u.u_error = ENFILE;
	return(NULL);
}
